use core::ops::Deref;
use core::ops::DerefMut;
use core::sync::atomic::AtomicUsize;
use alloc::boxed::Box;
use alloc::sync::Arc;
use alloc::vec::Vec;
use alloc::collections::BTreeMap;
use spin::Mutex;

use crate::config::PAGE_SIZE;
use crate::mm::address::PageNum;
use crate::mm::address::PhysAddr;
use crate::mm::kalloc::kalloc_page;
use super::BlockIOWrapper;
use super::BlockCache;

struct LRUItem {
    pub blockid: usize,
    pub prev: Option<usize>,
    pub next: Option<usize>,
}

struct LRUManager {
    blocksz_log: usize,
    maxblock: usize,

    head: Option<usize>,
    tail: Option<usize>,

    items: Vec<LRUItem>,
    pages: Vec<PageNum>,
    index: BTreeMap<usize, usize>,

    wrapper: Box<dyn BlockIOWrapper>,
}

impl LRUItem {
    pub fn new(blkid: usize, pre: Option<usize>, nxt: Option<usize>) -> Self {
        Self { blockid: blkid, prev: pre, next: nxt }
    }
}

impl LRUManager {
    fn blocksz(&self) -> usize {
        1 << self.blocksz_log
    }

    fn blocks_per_page(&self) -> usize {
        PAGE_SIZE >> self.blocksz_log
    }

    /// 返回表中第idx个LRUItem对应的不可变slice
    /// 如果还没分配页，则分配所需的页
    fn blk_phys(&mut self, idx: usize) -> PhysAddr {
        let bpp = self.blocks_per_page();
        let page_idx = idx / bpp;
        let page_off = idx % bpp;

        while self.pages.len() <= page_idx {
            self.pages.push(kalloc_page());
        }

        let phys_addr = self.pages[page_idx].offset_phys(page_off << self.blocksz_log);
        phys_addr
    }

    fn use_blk_existed(&mut self, idx: usize) {
        let head = self.head.unwrap();
        let tail = self.tail.unwrap();

        // LRUItem就在链表头，不用移动，直接返回
        if idx == head {
            return;
        }
        
        // 将idx对应的LRUItem移出链表，并处理好
        if idx == tail {
            // LRUItem在链表尾，且不是链表头
            // 此时需要：(1)处理前一个元素的指针 (2)使前一个元素变为self.tail
            let new_tail = self.items[idx].prev.unwrap();
            self.items[new_tail].next = None;
            self.tail = Some(new_tail);
        } else {
            // LRUItem在链表中间
            // 此时需要：处理前后指针
            let prev = self.items[idx].prev.unwrap();
            let next = self.items[idx].next.unwrap();
            self.items[prev].next = Some(next);
            self.items[next].prev = Some(prev);
        }

        // 将取出的idx元素变为LRU链表头
        let old_head = self.head.unwrap();
        self.items[old_head].prev = Some(idx);
        self.items[idx].prev = None;
        self.items[idx].next = Some(old_head);
            
        self.head = Some(idx);
    }

    fn use_blk_unexisted(&mut self, blkid: usize) {
        let item_cnt = self.items.len();

        if item_cnt < self.maxblock {
            // LRU链表还没达到最大大小，可以直添加元素
            self.items.push(LRUItem::new(blkid, None, self.head));
            self.index.insert(blkid, item_cnt);

            if self.head.is_none() {
                // 如果链表中还没有元素，则将self.head和self.tail指针赋值
                self.head = Some(item_cnt);
                self.tail = Some(item_cnt);
            } else {
                // 如果链表中已有元素，则只更新self.head，并处理指针
                let old_head = self.head.unwrap();
                self.items[old_head].prev = Some(item_cnt);
                self.head = Some(item_cnt);
            }
        } else {
            // LRU链表已经达到最大大小，需要替换
            let old_tail = self.tail.unwrap();
            let new_tail = self.items[old_tail].prev.unwrap();

            // 更新index
            let oldbid = self.items[old_tail].blockid;
            self.index.remove(&oldbid);
            self.index.insert(blkid, old_tail);

            self.items[new_tail].next = None;
            self.tail = Some(new_tail);

            let old_head = self.head.unwrap();
            self.items[old_tail].blockid = blkid;
            self.items[old_tail].next = Some(old_head);
            self.items[old_head].prev = Some(old_tail);
            self.items[old_tail].prev = None;
            self.head = Some(old_tail);
        }
    }
    
    /// 将blkid对应的item放在LRU的头部
    fn use_blk(&mut self, blkid: usize, do_read: bool) {
        match self.index.get(&blkid).cloned() {
            Some(idx) => self.use_blk_existed(idx),
            None => {
                self.use_blk_unexisted(blkid);
                if do_read {
                    let head = self.head.unwrap();
                    let mut blkaddr = self.blk_phys(head);
                    self.wrapper.read_block(blkid, blkaddr.as_slice_mut(self.blocksz()));
                }
            }
        }

    }

    fn read_block(&mut self, blkid: usize, buf: &mut [u8]) {
        self.use_blk(blkid, true);
        let head = self.head.unwrap();
        let mut blkaddr = self.blk_phys(head);
        buf.copy_from_slice(blkaddr.as_slice_mut(self.blocksz()));
    }

    fn write_block(&mut self, blkid: usize, buf: &[u8]) {
        self.use_blk(blkid, false);
        let head = self.head.unwrap();
        let mut blkaddr = self.blk_phys(head);
        blkaddr.as_slice_mut(self.blocksz()).copy_from_slice(buf);
        // write-through
        self.wrapper.write_block(blkid, buf);
    }
}

pub struct LRUCache(Mutex<LRUManager>);

unsafe impl Send for LRUCache { }
unsafe impl Sync for LRUCache { }

impl BlockCache for LRUCache {
    fn read_block(&self, blk: usize, buf: &mut[u8]) {
        let mut m = self.0.lock();
        m.read_block(blk, buf);
    }
    fn write_block(&self, blk: usize, buf: &[u8]) {
        let mut m = self.0.lock();
        m.write_block(blk, buf);
    }
    fn sync(&self) {

    }
}

impl LRUCache {
    pub fn new<T: BlockIOWrapper + 'static>(blksz_log: usize, maxblock: usize, wrp: T) -> Self {
        Self(Mutex::new(LRUManager {
            blocksz_log: blksz_log,
            maxblock: maxblock,
            head: None,
            tail: None,
            items: Vec::new(),
            pages: Vec::new(),
            index: BTreeMap::new(),
            wrapper: Box::new(wrp)
        }))
    }
}
