// From rcore-os/rCore-Tutorial-v3
// See https://github.com/rcore-os/rCore-Tutorial-v3/blob/main/LICENSE
// Distributed under GPLv3

#![allow(non_snake_case)]
pub mod abstraction;
pub mod platform;

use super::BlockDevice;
use abstraction::*;
use core::convert::TryInto;

pub struct SDCard<T> {
    spi: T,
    spi_cs: u32,
    is_hc: bool,
}

/*
 * Start Data tokens:
 *         Tokens (necessary because at nop/idle (and CS active) only 0xff is
 *         on the data/command line)
 */
/** Data token start byte, Start Single Block Read */
pub const SD_START_DATA_SINGLE_BLOCK_READ: u8 = 0xFE;
#[allow(unused)]
/** Data token start byte, Start Multiple Block Read */
pub const SD_START_DATA_MULTIPLE_BLOCK_READ: u8 = 0xFE;
/** Data token start byte, Start Single Block Write */
pub const SD_START_DATA_SINGLE_BLOCK_WRITE: u8 = 0xFE;
/** Data token start byte, Start Multiple Block Write */
pub const SD_START_DATA_MULTIPLE_BLOCK_WRITE: u8 = 0xFC;

pub const SD_STOP_DATA_MULTIPLE_BLOCK_WRITE: u8 = 0xFD;

pub const SEC_LEN: usize = 512;

/** SD commands */
#[repr(u8)]
#[derive(Debug, PartialEq, Eq, Copy, Clone)]
#[allow(unused)]
pub enum CMD {
    /** Software reset */
    CMD0 = 0,
    /** Check voltage range (SDC V2) */
    CMD8 = 8,
    /** Read CSD register */
    CMD9 = 9,
    /** Read CID register */
    CMD10 = 10,
    /** Stop to read data */
    CMD12 = 12,
    /** Change R/W block size */
    CMD16 = 16,
    /** Read block */
    CMD17 = 17,
    /** Read multiple blocks */
    CMD18 = 18,
    /** Number of blocks to erase (SDC) */
    ACMD23 = 23,
    /** Write a block */
    CMD24 = 24,
    /** Write multiple blocks */
    CMD25 = 25,
    /** Initiate initialization process (SDC) */
    ACMD41 = 41,
    /** Leading command for ACMD* */
    CMD55 = 55,
    /** Read OCR */
    CMD58 = 58,
    /** Enable/disable CRC check */
    CMD59 = 59,
    /** Custom CMD - just for testing */
    CMD1 = 0x1,
}

#[allow(unused)]
#[derive(Debug, Copy, Clone)]
pub enum InitError {
    CMDFailed(CMD, u8),
    CardCapacityStatusNotSet([u8; 4]),
    CannotGetCardInfo,
}

/**
 * Card Specific Data: CSD Register
 */
#[derive(Debug, Copy, Clone)]
pub struct SDCardCSD {
    pub CSDStruct: u8,        /* CSD structure */
    pub SysSpecVersion: u8,   /* System specification version */
    pub Reserved1: u8,        /* Reserved */
    pub TAAC: u8,             /* Data read access-time 1 */
    pub NSAC: u8,             /* Data read access-time 2 in CLK cycles */
    pub MaxBusClkFrec: u8,    /* Max. bus clock frequency */
    pub CardComdClasses: u16, /* Card command classes */
    pub RdBlockLen: u8,       /* Max. read data block length */
    pub PartBlockRead: u8,    /* Partial blocks for read allowed */
    pub WrBlockMisalign: u8,  /* Write block misalignment */
    pub RdBlockMisalign: u8,  /* Read block misalignment */
    pub DSRImpl: u8,          /* DSR implemented */
    pub Reserved2: u8,        /* Reserved */
    pub DeviceSize: u32,      /* Device Size */
    //MaxRdCurrentVDDMin: u8,   /* Max. read current @ VDD min */
    //MaxRdCurrentVDDMax: u8,   /* Max. read current @ VDD max */
    //MaxWrCurrentVDDMin: u8,   /* Max. write current @ VDD min */
    //MaxWrCurrentVDDMax: u8,   /* Max. write current @ VDD max */
    //DeviceSizeMul: u8,        /* Device size multiplier */
    pub EraseGrSize: u8,         /* Erase group size */
    pub EraseGrMul: u8,          /* Erase group size multiplier */
    pub WrProtectGrSize: u8,     /* Write protect group size */
    pub WrProtectGrEnable: u8,   /* Write protect group enable */
    pub ManDeflECC: u8,          /* Manufacturer default ECC */
    pub WrSpeedFact: u8,         /* Write speed factor */
    pub MaxWrBlockLen: u8,       /* Max. write data block length */
    pub WriteBlockPaPartial: u8, /* Partial blocks for write allowed */
    pub Reserved3: u8,           /* Reserded */
    pub ContentProtectAppli: u8, /* Content protection application */
    pub FileFormatGroup: u8,     /* File format group */
    pub CopyFlag: u8,            /* Copy flag (OTP) */
    pub PermWrProtect: u8,       /* Permanent write protection */
    pub TempWrProtect: u8,       /* Temporary write protection */
    pub FileFormat: u8,          /* File Format */
    pub ECC: u8,                 /* ECC code */
    pub CSD_CRC: u8,             /* CSD CRC */
    pub Reserved4: u8,           /* always 1*/
}

/**
 * Card Identification Data: CID Register
 */
#[derive(Debug, Copy, Clone)]
pub struct SDCardCID {
    pub ManufacturerID: u8, /* ManufacturerID */
    pub OEM_AppliID: u16,   /* OEM/Application ID */
    pub ProdName1: u32,     /* Product Name part1 */
    pub ProdName2: u8,      /* Product Name part2*/
    pub ProdRev: u8,        /* Product Revision */
    pub ProdSN: u32,        /* Product Serial Number */
    pub Reserved1: u8,      /* Reserved1 */
    pub ManufactDate: u16,  /* Manufacturing Date */
    pub CID_CRC: u8,        /* CID CRC */
    pub Reserved2: u8,      /* always 1 */
}

/**
 * Card information
 */
#[derive(Debug, Copy, Clone)]
pub struct SDCardInfo {
    pub SD_csd: SDCardCSD,
    pub SD_cid: SDCardCID,
    pub CardCapacity: u64,  /* Card Capacity */
    pub CardBlockSize: u64, /* Card Block Size */
}

impl<T: SPIActions> SDCard<T> {
  pub fn new(spi: T, spi_cs: u32) -> Self {
    Self {
      spi: spi,
      spi_cs: spi_cs,
      is_hc: false,
    }
  }

  fn HIGH_SPEED_ENABLE(&self) {
      self.spi.set_clk_rate(4);
  }

  fn CS_HIGH(&self) {
    self.spi.switch_cs(false, 0);
  }

  fn CS_LOW(&self) {
    self.spi.switch_cs(true, 0);
  }

  fn lowlevel_init(&self) {
    // gpiohs::set_direction(self.cs_gpionum, gpio::direction::OUTPUT);
    // at first clock rate shall be low (below 200khz)
    self.spi.init();
    self.spi.set_clk_rate(2500);
  }

  fn write_data(&self, data: &[u8]) {
    self.spi.configure(
      1,  // use lines
      8,  // bits per word
      true,  // endian: big-endian
    );
    self.spi.send_data(self.spi_cs, data);
  }
  
  fn read_data(&self, data: &mut [u8]) {
    self.spi.configure(
      1,  // use lines
      8,  // bits per word
      true,  // endian: big-endian
    );
    self.spi.recv_data(self.spi_cs, data);
  }

  /*
    * Send 5 bytes command to the SD card.
    * @param  cmd: The user expected command to send to SD card.
    * @param  arg: The command argument.
    * @param  crc: The CRC.
    * @retval None
    */
  fn send_cmd(&self, cmd: CMD, arg: u32, crc: u8) {
    self.CS_LOW();
    /* Send the Cmd bytes */
    self.write_data(&[
      /* Construct byte 1 */
      ((cmd as u8) | 0x40),
      /* Construct byte 2 */
      (arg >> 24) as u8,
      /* Construct byte 3 */
      ((arg >> 16) & 0xff) as u8,
      /* Construct byte 4 */
      ((arg >> 8) & 0xff) as u8,
      /* Construct byte 5 */
      (arg & 0xff) as u8,
      /* CRC */
      crc,
    ]);
  }

  /* Send end-command sequence to SD card */
  fn end_cmd(&self) {
    // TODO Does end operation need to deassert CS?
    self.CS_HIGH();
    /* Send the cmd byte */
    self.write_data(&[0xff; 4]);
  }

  /*
  * Returns the SD response.
  * @param  None
  * @retval The SD Response:
  *         - 0xFF: Sequence failed
  *         - 0: Sequence succeed
  */
  fn get_response(&self) -> u8 {
    let result = &mut [0xffu8];
    let mut timeout = 0xFFF;
    /* Check if response is got or a timeout is happen */
    while timeout != 0 {
      self.read_data(result);
      /* Right response got */
      if result[0] != 0xFF {
          return result[0];
      }
      timeout -= 1;
    }
    /* After time out */
    log!("sd":"get_response">"seq failed");
    return 0xFF;
  }

  /*
    * Get SD card data response.
    * @param  None
    * @retval The SD status: Read data response xxx0<status>1
    *         - status 010: Data accecpted
    *         - status 101: Data rejected due to a crc error
    *         - status 110: Data rejected due to a Write error.
    *         - status 111: Data rejected due to other error.
    */
  fn get_dataresponse(&self) -> u8 {
    let response = &mut [0u8];
    /* Read resonse */
    self.read_data(response);
    /* Mask unused bits */
    response[0] &= 0x1F;
    if response[0] != 0x05 {
      return 0xFF;
    }
    /* Wait null data */
    self.read_data(response);
    while response[0] == 0 {
      self.read_data(response);
    }
    /* Return response */
    return 0;
  }

  /*
  * Read the CSD card register
  *         Reading the contents of the CSD register in SPI mode is a simple
  *         read-block transaction.
  * @param  SD_csd: pointer on an SCD register structure
  * @retval The SD Response:
  *         - `Err()`: Sequence failed
  *         - `Ok(info)`: Sequence succeed
  */
  fn get_csdregister(&self) -> Result<SDCardCSD, ()> {
    let mut csd_tab = [0u8; 18];
    /* Send CMD9 (CSD register) */
    log!("sd":"csd">"sending cmd 9");
    self.send_cmd(CMD::CMD9, 0, 0);
    /* Wait for response in the R1 format (0x00 is no errors) */
    let resp = self.get_response();
    if resp != 0x00 {
      log!("sd":"csd">"cmd9 response error: 0x{:x}", resp);
      self.end_cmd();
      return Err(());
    }

    let resp = self.get_response();
    if resp != SD_START_DATA_SINGLE_BLOCK_READ {
      log!("sd":"csd">"cmd9 get_data error: 0x{:x}", resp);
      self.end_cmd();
      return Err(());
    }
    /* Store CSD register value on csd_tab */
    /* Get CRC bytes (not really needed by us, but required by SD) */
    self.read_data(&mut csd_tab);
    self.end_cmd();
    self.end_cmd();
    /* see also: https://cdn-shop.adafruit.com/datasheets/TS16GUSDHC6.pdf */
    return Ok(SDCardCSD {
      /* Byte 0 */
      CSDStruct: (csd_tab[0] & 0xC0) >> 6,
      SysSpecVersion: (csd_tab[0] & 0x3C) >> 2,
      Reserved1: csd_tab[0] & 0x03,
      /* Byte 1 */
      TAAC: csd_tab[1],
      /* Byte 2 */
      NSAC: csd_tab[2],
      /* Byte 3 */
      MaxBusClkFrec: csd_tab[3],
      /* Byte 4, 5 */
      CardComdClasses: (u16::from(csd_tab[4]) << 4) | ((u16::from(csd_tab[5]) & 0xF0) >> 4),
      /* Byte 5 */
      RdBlockLen: csd_tab[5] & 0x0F,
      /* Byte 6 */
      PartBlockRead: (csd_tab[6] & 0x80) >> 7,
      WrBlockMisalign: (csd_tab[6] & 0x40) >> 6,
      RdBlockMisalign: (csd_tab[6] & 0x20) >> 5,
      DSRImpl: (csd_tab[6] & 0x10) >> 4,
      Reserved2: 0,
      // DeviceSize: (csd_tab[6] & 0x03) << 10,
      /* Byte 7, 8, 9 */
      DeviceSize: ((u32::from(csd_tab[7]) & 0x3F) << 16)
          | (u32::from(csd_tab[8]) << 8)
          | u32::from(csd_tab[9]),
      /* Byte 10 */
      EraseGrSize: (csd_tab[10] & 0x40) >> 6,
      /* Byte 10, 11 */
      EraseGrMul: ((csd_tab[10] & 0x3F) << 1) | ((csd_tab[11] & 0x80) >> 7),
      /* Byte 11 */
      WrProtectGrSize: (csd_tab[11] & 0x7F),
      /* Byte 12 */
      WrProtectGrEnable: (csd_tab[12] & 0x80) >> 7,
      ManDeflECC: (csd_tab[12] & 0x60) >> 5,
      WrSpeedFact: (csd_tab[12] & 0x1C) >> 2,
      /* Byte 12,13 */
      MaxWrBlockLen: ((csd_tab[12] & 0x03) << 2) | ((csd_tab[13] & 0xC0) >> 6),
      /* Byte 13 */
      WriteBlockPaPartial: (csd_tab[13] & 0x20) >> 5,
      Reserved3: 0,
      ContentProtectAppli: (csd_tab[13] & 0x01),
      /* Byte 14 */
      FileFormatGroup: (csd_tab[14] & 0x80) >> 7,
      CopyFlag: (csd_tab[14] & 0x40) >> 6,
      PermWrProtect: (csd_tab[14] & 0x20) >> 5,
      TempWrProtect: (csd_tab[14] & 0x10) >> 4,
      FileFormat: (csd_tab[14] & 0x0C) >> 2,
      ECC: (csd_tab[14] & 0x03),
      /* Byte 15 */
      CSD_CRC: (csd_tab[15] & 0xFE) >> 1,
      Reserved4: 1,
      /* Return the reponse */
    });
  }

  /*
  * Read the CID card register.
  *         Reading the contents of the CID register in SPI mode is a simple
  *         read-block transaction.
  * @param  SD_cid: pointer on an CID register structure
  * @retval The SD Response:
  *         - `Err()`: Sequence failed
  *         - `Ok(info)`: Sequence succeed
  */
  fn get_cidregister(&self) -> Result<SDCardCID, ()> {
    let mut cid_tab = [0u8; 18];
    /* Send CMD10 (CID register) */
    log!("sd":"csd">"sending cmd 10");
    self.send_cmd(CMD::CMD10, 0, 0);
    /* Wait for response in the R1 format (0x00 is no errors) */
    let resp = self.get_response();
    if resp != 0x00 {
      log!("sd":"csd">"sending cmd 10: 0x{:x}", resp);
      self.end_cmd();
      return Err(());
    }

    let resp = self.get_response();
    if resp != SD_START_DATA_SINGLE_BLOCK_READ {
      log!("sd":"csd">"cmd 10 get_data error: 0x{:x}", resp);
      self.end_cmd();
      return Err(());
    }
    /* Store CID register value on cid_tab */
    /* Get CRC bytes (not really needed by us, but required by SD) */
    self.read_data(&mut cid_tab);
    self.end_cmd();
    return Ok(SDCardCID {
      /* Byte 0 */
      ManufacturerID: cid_tab[0],
      /* Byte 1, 2 */
      OEM_AppliID: (u16::from(cid_tab[1]) << 8) | u16::from(cid_tab[2]),
      /* Byte 3, 4, 5, 6 */
      ProdName1: (u32::from(cid_tab[3]) << 24)
          | (u32::from(cid_tab[4]) << 16)
          | (u32::from(cid_tab[5]) << 8)
          | u32::from(cid_tab[6]),
      /* Byte 7 */
      ProdName2: cid_tab[7],
      /* Byte 8 */
      ProdRev: cid_tab[8],
      /* Byte 9, 10, 11, 12 */
      ProdSN: (u32::from(cid_tab[9]) << 24)
          | (u32::from(cid_tab[10]) << 16)
          | (u32::from(cid_tab[11]) << 8)
          | u32::from(cid_tab[12]),
      /* Byte 13, 14 */
      Reserved1: (cid_tab[13] & 0xF0) >> 4,
      ManufactDate: ((u16::from(cid_tab[13]) & 0x0F) << 8) | u16::from(cid_tab[14]),
      /* Byte 15 */
      CID_CRC: (cid_tab[15] & 0xFE) >> 1,
      Reserved2: 1,
    });
  }

  /*
    * Returns information about specific card.
    * @param  cardinfo: pointer to a SD_CardInfo structure that contains all SD
    *         card information.
    * @retval The SD Response:
    *         - `Err(())`: Sequence failed
    *         - `Ok(info)`: Sequence succeed
    */
  fn get_cardinfo(&self) -> Result<SDCardInfo, ()> {
    let mut info = SDCardInfo {
      SD_csd: self.get_csdregister()?,
      SD_cid: self.get_cidregister()?,
      CardCapacity: 0,
      CardBlockSize: 0,
    };
    info.CardBlockSize = 1 << u64::from(info.SD_csd.RdBlockLen);
    info.CardCapacity = (u64::from(info.SD_csd.DeviceSize) + 1) * 1024 * info.CardBlockSize;

    Ok(info)
  }

  #[allow(unused)]
  fn retry_cmd(&self, cmd: CMD, arg: u32, crc: u8, retry_times: u32, success_cond: &dyn Fn(u8)->bool) -> Result<u8, u8> {
    let mut resp = self.get_response();

    for i in 0..retry_times {
      log!("sd":"retry_cmd">"{:?} retried {} time", cmd, i);
      self.send_cmd(cmd, arg, crc);
      resp = self.get_response();
      self.end_cmd();
      if success_cond(resp) {
        return Ok(resp);
      }
    }
    
    log!("sd":"retry_cmd">"{:?} responded with code {:#x}", cmd, resp);
    return Err(resp);
  }

  /*
    * Initializes the SD/SD communication in SPI mode.
    * @param  None
    * @retval The SD Response info if succeeeded, otherwise Err
    */
  pub fn init(&mut self) -> Result<SDCardInfo, InitError> {
    log!("sd":"init">"spi lowlevel init");
    /* Initialize SD_SPI */
    self.lowlevel_init();
    /* An empty frame for commands */
    let mut frame = [0u8; 4];
    /* NOTE: this reset doesn't always seem to work if the SD access was broken off in the
      * middle of an operation: CMDFailed(CMD0, 127). */

    /* Send dummy byte 0xFF, 10 times with CS high */
    /* Rise CS and MOSI for 80 clocks cycles */
    /* Send dummy byte 0xFF */
    log!("sd":"init">"switch cs");
    self.spi.switch_cs(false, 0);
    log!("sd":"init">"configure");
    self.spi.configure(
      1,  // use lines
      8,  // bits per word
      true,  // endian: big-endian
    );
    log!("sd":"init">"write 20 0xff");
    self.write_data(&[0xff; 20]);
    /*------------Put SD in SPI mode--------------*/
    /* SD initialized and set to SPI mode properly */

    /* Send software reset */
    let mut result;

    log!("sd":"init">"send cmd0");
    if let Err(resp) = self.retry_cmd(CMD::CMD0, 0, 0x95, 200, &|x|->bool { x==0x01 }) {
      return Err(InitError::CMDFailed(CMD::CMD0, 0));
    }

    log!("sd":"init">"send cmd8");
    result = match self.retry_cmd(CMD::CMD8, 0x01AA, 0x87, 2, &|x|->bool { x==0x01 || x==0x05 }) {
      Ok(resp)  => { resp },
      Err(resp) => { return Err(InitError::CMDFailed(CMD::CMD8, resp)); }
    };

    if result != 0x01 {
      // Standard Capacity Card
      log!("sd":"init">"std");
      result = 0xff;

      while result != 0x00 {
        log!("sd":"init">"send cmd55");
        self.send_cmd(CMD::CMD55, 0, 0);
        log!("sd":"init">"get_response cmd55");
        self.get_response();
        log!("sd":"init">"end cmd55");
        self.end_cmd();
        log!("sd":"init">"send acmd41");
        self.send_cmd(CMD::ACMD41, 0x40000000, 0);
        log!("sd":"init">"get_response acmd41");
        result = self.get_response();
        log!("sd":"init">"end acmd41");
        self.end_cmd();
      }

    } else {
      // Need further discrimination
      log!("sd":"init">"discriminate");
      let mut cnt = 0;

      cnt = 0xff;
      log!("sd":"init">"cmd55 and 41");
      while result != 0x00 {
        log!("sd":"init">"send cmd55");
        self.send_cmd(CMD::CMD55, 0, 0);
        log!("sd":"init">"get_response cmd55");
        self.get_response();
        log!("sd":"init">"end cmd55");
        self.end_cmd();
        log!("sd":"init">"send cmd41");
        self.send_cmd(CMD::ACMD41, 0x40000000, 0);
        log!("sd":"init">"get_response cmd41");
        result = self.get_response();
        log!("sd":"init">"end cmd41 with response 0x{:x}", result);
        self.end_cmd();

        if cnt == 0 {
          return Err(InitError::CMDFailed(CMD::ACMD41, 0));
        } else {
          cnt -= 1;
        }
      }
      
      cnt = 0xff;
      result = 0xff;
      log!("sd":"init">"cmd58");
      while (result != 0x0) && (result != 0x1) {
        log!("sd":"init">"send cmd58");
        self.send_cmd(CMD::CMD58, 0, 1);
        log!("sd":"init">"get_response cmd58");
        result = self.get_response();
        log!("sd":"init">"cmd58 read ocr");
        self.read_data(&mut frame);
        log!("sd":"init">"cmd58 end cmd");
        self.end_cmd();

        if cnt == 0 {
          return Err(InitError::CMDFailed(CMD::CMD58, 0));
        } else {
          cnt -= 1;
        }
      }

      for i in 0..4 {
        log!("sd":"init">"OCR[{}]: 0x{:x}", i, frame[i]);
      }

      if (frame[0] & 0x40) != 0 {
        log!("sd":"init">"HC card!");
        self.is_hc = true;
      } else {
        log!("sd":"init">"SC card!");
        self.is_hc = false;
      }
    }

    self.spi.switch_cs(false, 0);
    self.write_data(&[0xff; 10]);

    self.HIGH_SPEED_ENABLE();
    match self.get_cardinfo() {
      Ok(info)     => { Ok(info) }
      Err(_)       => { Err(InitError::CannotGetCardInfo) }
    }
  }


  /*
  * Reads a block of data from te SD.
  * @param  data_buf: slice that receives the data read from the SD.
  * @param  sector: SD's internal address to read from.
  * @retval The SD Response:
  *         - `Err(())`: Sequence failed
  *         - `Ok(())`: Sequence succeed
  */
  pub fn read_sector(&self, data_buf: &mut [u8], sector: u32) -> Result<(), ()> {
    assert!(data_buf.len() >= SEC_LEN && (data_buf.len() % SEC_LEN) == 0);
    let sector: u32 = match self.is_hc {
      false => { sector << 9 },
      _ =>     { sector }
    };

    let mut cur_sector = sector;
    
    for chunk in data_buf.chunks_mut(SEC_LEN) {

      log!("sd":"read">"reading sector {}", cur_sector);
      
      /* Send CMD17 to read one block */
      let mut count = 200;
      loop {
        self.send_cmd(CMD::CMD17, cur_sector, 0);
        /* Check if the SD acknowledged the read block command: R1 response (0x00: no errors) */
        let res = self.get_response();
        
        if count == 0 {
          log!("sd":"read">"cmd 17 returned error code {:#x}", res);
          self.end_cmd();
          return Err(());
        }

        if res == 0x0 {
          break;
        }
        
        self.end_cmd();
        count -= 1;
      }

      //let mut dma_chunk = [0u32; SEC_LEN];
      let mut tmp_chunk= [0u8; SEC_LEN];
      let mut resp = self.get_response();
      if resp != SD_START_DATA_SINGLE_BLOCK_READ {
        log!("sd":"read">"sd responding {:#x} rather than 0xfe", resp);
        return Err(());
      }

      /* Read the SD block data : read NumByteToRead data */
      self.read_data(&mut tmp_chunk);
      /* Place the data received as u32 units from DMA into the u8 target buffer */
      for (a, b) in chunk.iter_mut().zip(/*dma_chunk*/tmp_chunk.iter()) {
        *a = *b;
      }
      /* Get CRC bytes (not really needed by us, but required by SD) */
      let mut frame = [0u8; 2];
      self.read_data(&mut frame);

      // put some dummy bytes
      self.end_cmd();
      self.end_cmd();

      cur_sector += match self.is_hc {
        false => { SEC_LEN as u32 },
        _     => { 1 }
      };
    }

    // put some dummy bytes
    self.end_cmd();
    self.end_cmd();

    log!("sd":"read">"exit read");

    /* It is an error if not everything requested was read */
    Ok(())
  }

  /*
    * Writes a block to the SD
    * @param  data_buf: slice containing the data to be written to the SD.
    * @param  sector: address to write on.
    * @retval The SD Response:
    *         - `Err(())`: Sequence failed
    *         - `Ok(())`: Sequence succeed
    */
  pub fn write_sector(&self, data_buf: &[u8], sector: u32) -> Result<(), ()> {
    assert!(data_buf.len() >= SEC_LEN && (data_buf.len() % SEC_LEN) == 0);
    let sector: u32 = match self.is_hc {
      false => { sector << 9 },
      _ =>     { sector }
    };

    let mut cur_sector = sector;

    let frame = [0xff, SD_START_DATA_SINGLE_BLOCK_WRITE];

    for trunk in data_buf.chunks(SEC_LEN) {

      let mut count = 200;
      loop {
        self.send_cmd(CMD::CMD24, cur_sector, 0);
        /* Check if the SD acknowledged the read block command: R1 response (0x00: no errors) */
        let res = self.get_response();
        
        if count == 0 {
          log!("sd":"read">"cmd 24 returned error code {:#x}", res);
          self.end_cmd();
          return Err(());
        }

        if res == 0x0 {
          log!("test_cannot_run_mmap_munmap":"write_sec">"cmd24 resp: {}", res);
          break;
        }
        
        count -= 1;
      }

      self.write_data(&frame);
      self.write_data(trunk);
      self.write_data(&[0xff, 0xff]);  // dummy crc

      let mut resp = self.get_response() & 0x1F;
      if resp != 0x5 {
        log!("test_cannot_run_mmap_munmap":"write_sec">"sec: {:#x}, resp: {}", cur_sector, resp);
      }

      while self.get_response() != 0xff { }

      self.end_cmd();
      self.end_cmd();
      
      cur_sector += match self.is_hc {
        false => { SEC_LEN as u32 },
        _     => { 1 }
      };
    }

    self.end_cmd();
    self.end_cmd();

    Ok(())
  }
}

/** CS value passed to SPI controller, this is a dummy value as SPI0_CS3 is not mapping to anything
 * in the FPIOA */
const SD_CS: u32 = 0;

pub fn init_sdcard() -> SDCard<SPIImpl> {
  // wait previous output
  for _ in 0..1000000 { }

  let spi = SPIImpl::new(abstraction::SPIDevice::QSPI2);
  let mut sd = SDCard::new(spi, SD_CS);
  let info = sd.init().unwrap();
  // assert!(num_sectors > 0);

  println!("[kernel] init sdcard!");
  sd
}

pub struct SDCardWrapper(SDCard<SPIImpl>);

impl SDCardWrapper {
  pub fn new() -> Self {
    Self(init_sdcard())
  }

  pub fn init(&self) { }
}

impl BlockDevice for SDCardWrapper {
  fn read_block(&self, block_id: usize, buf: &mut [u8]) {
    if let Err(_) = self.0.read_sector(buf, block_id as u32) {
        panic!("read_block invalid {}", block_id);
    }
  }
  fn write_block(&self, block_id: usize, buf: &[u8]) {
    if let Err(_) = self.0.write_sector(buf, block_id as u32) {
        panic!("write_block invalid {}", block_id);
    }
  }
}
