use fat32::file;
use fat32::dir;
use fat32::fat;

use alloc::sync::Arc;
use crate::vfs::*;
use spin::Lazy;

static SDCARD: Lazy<SDCard> = Lazy::new(|| { SDCard { } });
static VOLUMN: Lazy<fat32::volume::Volume<SDCard>> = Lazy::new(|| { fat32::volume::Volume::new(*SDCARD)});
pub static FAT32ROOT: Lazy<Inode> = Lazy::new(|| { Arc::new(VOLUMN.root_dir()) });

#[derive(Copy, Clone)]
struct SDCard {}

impl block_device::BlockDevice for SDCard {
    type Error = ();
    fn read(&self, buf: &mut[u8], address: usize, _number_of_blocks: usize) -> Result<(), Self::Error> {
        #[cfg(feature = "fat_offset")]
        let address = address + 512 * 10274;
        let mut blk = address / 512;
        let mut len_left = buf.len();
        let mut blk_off = address % 512;
        let mut tmp_buf: [u8; 512] = [0; 512];
        let total_len = buf.len();

        while len_left > 0 {
            crate::blockdev::read_block(blk, &mut tmp_buf);
            let single_len = core::cmp::min(len_left, 512 - blk_off);
            buf[total_len - len_left..(total_len - len_left + single_len)]
                .copy_from_slice(&tmp_buf[blk_off..blk_off+single_len]);

            len_left -= single_len;
            blk += 1;
            blk_off = (blk_off + single_len) % 512;
        }
        Ok(())
    }
    fn write(&self, buf: &[u8], address: usize, _number_of_blocks: usize) -> Result<(), Self::Error> {
        #[cfg(feature = "fat_offset")]
        let address = address + 512 * 10274;
        let mut blk = address / 512;
        let mut len_left = buf.len();
        let mut blk_off = address % 512;
        let mut tmp_buf: [u8; 512] = [0; 512];
        let total_len = buf.len();

        while len_left > 0 {
            let single_len = core::cmp::min(len_left, 512 - blk_off);
            if single_len != 512 {
                crate::blockdev::read_block(blk, &mut tmp_buf);
            }
            tmp_buf[blk_off..blk_off+single_len]
                .copy_from_slice(&buf[total_len - len_left..(total_len - len_left + single_len)]);

            crate::blockdev::write_block(blk, &mut tmp_buf);

            len_left -= single_len;
            blk += 1;
            blk_off = (blk_off + single_len) % 512;
        }
        Ok(())
    }
}


#[allow(unused)]
pub fn sd_test() {
    let mut root = VOLUMN.root_dir();
    // root.create_dir("from_root_dir").unwrap();
    root.direntry_iter().map(|dent| {
        if dent.is_lfn() {
            let (buf, len) = dent.get_lfn().unwrap();
            println!("Dent: {}", unsafe { core::str::from_utf8_unchecked(&buf[..len]) });
        } else {
            let (buf, len) = dent.get_sfn().unwrap();
            println!("Dent: {}", unsafe { core::str::from_utf8_unchecked(&buf[..len]) });
        }
    }).count();
    /*
    ROOT.unlink_child("hello.txt", false).unwrap_or(0);
    let inode = ROOT.create("hello.txt", FileMode::all(), InodeType::File).unwrap();
    let mut write_buf: [u8; 1025] = [0; 1025];
    for i in 0..write_buf.len() {
        write_buf[i] = 'A' as u8 + (i % 26) as u8;
    }
    let write_len = inode.write_offset(0, &write_buf).unwrap();
    println!("write lenght {}", write_len);
    let mut buf: [u8; 1] = [69];
    println!("inode lenght {}", inode.len());
    for i in 0..inode.len() {
        inode.read_offset(i, &mut buf).unwrap();
        print!("{}", buf[0] as char);
    }
    */
}

impl From<fat32::dir::DirError> for FileErr {
    fn from(err: fat32::dir::DirError) -> FileErr {
        match err {
            fat32::dir::DirError::DirHasExist => {
                FileErr::InodeChildExist
            }
            fat32::dir::DirError::FileHasExist => {
                FileErr::InodeChildExist
            }
            fat32::dir::DirError::IllegalChar => {
                FileErr::InodeNotChild
            }
            _ => {
                FileErr::NotDefine
            }
        }
    }
}

impl crate::vfs::_Inode for fat32::file::File<'_, SDCard> {
    fn read_offset(&self, offset: usize, buf: &mut [u8]) -> Result<usize, FileErr> {
        match self.read_off(offset, buf) {
            Err(fat32::file::FileError::WriteError) => {
                Err(FileErr::NotDefine)
            }
            Err(fat32::file::FileError::BufTooSmall) => {
                Err(FileErr::NotDefine)
            }
            Ok(len) => {
                Ok(len)
            }
        }
    }

    fn write_offset(&self, offset: usize, buf: &[u8]) -> Result<usize, FileErr> {
        let _self = unsafe { (self as *const Self as *mut Self).as_mut().unwrap()};
        match _self.write_off(offset, buf) {
            Err(fat32::file::FileError::WriteError) => {
                Err(FileErr::NotDefine)
            }
            Err(fat32::file::FileError::BufTooSmall) => {
                Err(FileErr::NotDefine)
            }
            Ok(len) => {
                Ok(len)
            }
        }
    }

    fn len(&self) -> usize {
        self.detail.length().unwrap()
    }

    fn ino(&self) -> Result<Ino, FileErr> {
        Ok(self.detail.cluster() as usize)
    }
}

use core::convert::TryFrom;
use core::convert::TryInto;
use fat32::entry::Entry;
use fat32::dir::Dir;

fn inode_try_from_entry(di: Entry, parent: &Dir<'static, SDCard>) -> Result<Inode, FileErr> {
    if di.is_file() {
        let fat = fat::FAT::new(di.cluster(),
                            parent.device,
                            parent.bpb.fat1());
        log!("fs":"get_child">"NEW a fat struct");
        Ok(Arc::new(file::File::<SDCard> {
            device: parent.device,
            bpb: parent.bpb,
            dir_cluster: { 
                let cluster = parent.detail.cluster();
                log!("fs":"get_child">"get detail cluster is {}", cluster);
                cluster
            },
            detail: di,
            fat,
        }))
    } else if di.is_dir() {
        let fat = fat::FAT::new(di.cluster(),
                            parent.device,
                            parent.bpb.fat1());
        Ok(Arc::new(dir::Dir::<SDCard> {
            device: parent.device,
            bpb: parent.bpb,
            detail: di,
            fat,
        }))
    } else {
        log!("fs":"inode_try_from_entry">"error unknown type");
        Err(FileErr::UnknownType)
    }

}

impl crate::vfs::_Inode for fat32::dir::Dir<'static, SDCard> {
    fn get_child(&self, name: &str) -> Result<Inode, FileErr> {
        match self.exist(name) {
            None => { 
                log!("fs":"get_child">"not found child {}", name);
                Err(FileErr::InodeNotChild)
            }
            Some(di) => {
                log!("fs":"get_child">"found child {}", name);
                inode_try_from_entry(di, self)
            }
        }
    }

    fn get_dirent(&self, offset: usize, dirent: &mut LinuxDirent) -> Result<usize, FileErr> {
        // fat32 目录项是32字节
        if offset % 32 != 0 {
            log!("fs":"get_dirent">"offset % 32 != 0");
            return Err(FileErr::NotDefine)
        }
        let dent_iter = self.direntry_iter().skip(offset / 32);

        let mut skip = 0;
        let dent = dent_iter.skip_while(|di| {
            if di.is_deleted() {
                skip += 1;
                true
            } else {
                false
            }
        }).take(1).last();

        if let None = dent {
            log!("fs":"get_dirent">"empty dir");
            return Err(FileErr::InodeEndOfDir)
        }

        let dent = dent.unwrap();
        if dent.is_lfn() {
            let (buf, len) = dent.get_lfn().unwrap();
            dirent.d_name[0..len].copy_from_slice(&buf[0..len]);
            dirent.d_ino = 0;
        } else {
            let (buf, len) = dent.get_sfn().unwrap();
            dirent.d_name[0..len].copy_from_slice(&buf[0..len]);
            dirent.d_ino = dent.cluster() as usize;
        }

        dirent.d_type = if dent.is_dir() {
                log!("fs":"get_dirent">"child type: Dir, cluster is {}", dirent.d_ino);
                DT_DIR
            } else if dent.is_file() {
                log!("fs":"get_dirent">"child type: File, cluster is {}", dirent.d_ino);
                DT_REG
            } else {
                log!("fs":"get_dirent">"child type: Unknown, cluster is {}", dirent.d_ino);
                DT_UNKNOWN
            };
        dirent.d_reclen = match u16::try_from(core::mem::size_of::<LinuxDirent>()) {
            Ok(size) => size,
            Err(_) => {
                log!("fs":"get_dirents">"invalid reclen");
                return Err(FileErr::NotDefine);
            }
        };
        dirent.d_off = match isize::try_from(offset + 1) {
            Ok(size) => size,
            Err(_) => {
                log!("fs":"get_dirents">"invalid d_off");
                return Err(FileErr::NotDefine);
            }
        };
        Ok(32 + skip * 32)
    }

    fn create(&self, name: &str, _: FileMode, itype: InodeType) -> Result<Inode, FileErr> {
        log!("fs":"create">"name: {}, type: {:?}", name, itype);
        let _self = unsafe { (self as *const Self as *mut Self).as_mut().unwrap()};
        match itype {
            InodeType::Directory => {
                log!("fs":"create">"Directory name: {}", name);
                let di = _self.create_dir(name)?;
                log!("fs":"create">"successed Directory name: {}", name);
                inode_try_from_entry(di, self)
            }
            InodeType::File => {
                log!("fs":"create">"File name: {}", name);
                let di = _self.create_file(name)?;
                log!("fs":"create">"successed File name: {}", name);
                inode_try_from_entry(di, self)
            }
            _ => {
                log!("fs":"create">"Not support type {:?}", itype);
                Err(FileErr::NotDefine)
            }
        }
    }

    fn unlink_child(&self, name: &str, rm_dir: bool) -> Result<usize, FileErr> {
        let _self = unsafe { (self as *const Self as *mut Self).as_mut().unwrap()};
        match _self.delete_file(name) {
            Err(fat32::dir::DirError::NoMatchFile) => {
            }
            Err(e) => {
                return Err(e.into())
            }
            Ok(_) => {
                return Ok(0)
            }
        }
        if rm_dir {
            _self.delete_dir(name)?;
            return Ok(0)
        } else {
            return Err(FileErr::NotDefine)
        }
    }

    fn len(&self) -> usize {
        0
    }

    fn ino(&self) -> Result<Ino, FileErr> {
        Ok(self.detail.cluster() as usize)
    }
}

#[allow(unused)]
pub fn fat32_test() {
    assert!(FAT32ROOT.get_child("test_mmap.txt").is_err());
    assert!(FAT32ROOT.get_child("djifaosdjfioio.txt").is_err());
    assert!(FAT32ROOT.get_child("mmap").is_ok());
    assert!(FAT32ROOT.get_child("gettimeofday").is_ok());
    let mnt = FAT32ROOT.get_child("mnt").unwrap();
    let _ = mnt.get_child("test_mount").unwrap();

    let inode = FAT32ROOT.create("fat_test", FileMode::empty(), InodeType::File).unwrap();
    let buf: [u8; 10] = [1; 10];
    assert!(inode.write_offset(0, &buf).unwrap() == 10);

    let inode = FAT32ROOT.get_child("fat_test").unwrap();
    let mut buf: [u8; 10] = [0; 10];
    assert!(inode.read_offset(0, &mut buf).unwrap() == 10);
    for i in 0..10 {
        assert!(buf[i] == 1);
    }
    panic!()
}
