// [no_std] Don't use standard library
#![no_std]
// [no_main] Tell compiler we don't need initialization before main() #![no_main]
#![no_main]
#![feature(naked_functions)]
// [global_asm] allow include an assemble file
#![feature(panic_info_message)]
#![feature(alloc_error_handler)]
#![feature(ptr_to_from_bits)]
#![feature(const_trait_impl)]

use crate::process::cpu::{hart_enable_timer_interrupt, init_hart};
use core::{arch::asm, convert::TryInto};

#[macro_use]
mod lang_items;
mod sbi;

#[macro_use]
mod console;

mod bcache;
mod blockdev;
mod entry;
mod heap;
mod info;
mod link_syms;
mod mm;
mod process;
mod task;
mod trap;
mod user;

mod clock;
mod config;
mod vfs;
mod fat32_wrapper;
mod fpu;

extern crate alloc;
extern crate buddy_system_allocator;
extern crate spin;
#[macro_use]
extern crate bitflags;
extern crate elf_parser;
extern crate fat32;
extern crate block_device;

use mm::*;
use process::cpu::init_hartid;
use task::*;

fn pressure_test(block: u32, retry_times: u32) {
    let mut first = [0x0u8; 512];
    let mut trail = [0x0u8; 512];

    blockdev::read_block(block as usize, &mut first);
    for _ in 0..retry_times {
        blockdev::read_block(block as usize, &mut trail);
        for i in 0..512 {
            assert!(first[i] == trail[i]);
        }
    }
}

/// Clear .bss section
fn clear_bss() {
    (link_syms::sbss as usize..link_syms::ebss as usize)
        .for_each(|a| unsafe { (a as *mut u8).write_volatile(0) });
}

// 记录启动核
static mut BOOTHART: isize = -1;

// [no_mangle] Turn off Rust's name mangling
#[no_mangle]
extern "C" fn kernel_start() {
    let hart_id = init_hartid();

    if unsafe { BOOTHART } == -1 {
        unsafe {
            BOOTHART = hart_id.try_into().unwrap();
        };

        clear_bss();

        // 需要在开启虚拟内存之前初始化时钟，
        // 因为内核不会映射时钟配置寄存器
        #[cfg(feature = "init_clock")]
        clock::clock_init();

        heap::init();

        mm::init();

        fpu::init();

        init_hart(hart_id);

        // vfs::devfs::init_devfs().unwrap();

        // fat32_wrapper::fat32_test();

        // fat32_wrapper::sd_test();
        /*
        let mut first_read_buf = [0x0u8; 512];
        blockdev::read_block(0, &mut first_read_buf);
        for i in 0..512 {
            print!("{:#x} ", first_read_buf[i]);
            if (i&0xf) == 0xf {
                println!("");
            }
        }

        for i in 0..1000 {
            pressure_test(i as u32, 1);
        }
        */

        // Load shell
        #[cfg(feature = "load_contest_test")]
        scheduler_load_pcb(MemorySpace::from_elf(&|off: usize, buf: &mut[u8]| {
            buf.copy_from_slice(&user::RUNBUSYBOX[off..off+buf.len()]); Ok(())
        }).unwrap().0);

        #[cfg(feature = "load_shell")]
        scheduler_load_pcb(MemorySpace::from_elf(&|off: usize, buf: &mut[u8]| {
            buf.copy_from_slice(&user::SHELL[off..off+buf.len()]); Ok(())
        }).unwrap().0);

        #[cfg(feature = "batch")]
        for i in user::BATCH.iter() {
            scheduler_load_pcb(MemorySpace::from_elf(&|off: usize, buf: &mut[u8]| {
                buf.copy_from_slice(&i[off..off+buf.len()]); Ok(())
            }).unwrap().0);
        }

        #[cfg(feature = "multicore")]
        for i in 1..=4 {
            if hartid() != i {
                sbi::sbi_hsm_hart_start(i, crate::link_syms::skernel as usize, 0);
            }
        }
    } else {
        init_hart(hart_id);
    }
    trap::init();
    #[cfg(feature = "time_interrupt")]
    hart_enable_timer_interrupt();
    schedule();
}
