pub mod syscall;
use core::arch::global_asm;

use riscv::register::{
    mtvec::TrapMode,
    scause::{self, Exception, Interrupt, Trap},
    stval, stvec,
};

use crate::mm::*;
use crate::process::cpu::*;
use crate::task::*;

extern "C" {
    pub fn __alltraps();
    pub fn __restore(cx: usize);
}

global_asm!(include_str!("traps.s"));

pub fn init() {
    unsafe {
        stvec::write(__alltraps as usize, TrapMode::Direct);
    }
}

pub extern "C" fn trap_handler() {
    // Fixme: Don't skip the reference lifetime checker;
    current_pcb()
        .unwrap()
        .lock()
        .utimes_add(get_time() - current_hart_set_trap_times(get_time()));
    let scause = scause::read();
    let stval = stval::read();
    if let Trap::Exception(Exception::InstructionPageFault) = scause.cause() {
        log!("trap":"handler">"entry");
    }
    // println!("scause {:?}, stval 0x{:x}, sepc 0x{:x}", scause.cause(), stval, riscv::register::sepc::read());
    match scause.cause() {
        Trap::Exception(Exception::UserEnvCall) => {
            syscall::syscall_handler();
        }
        Trap::Exception(Exception::StoreFault) | Trap::Exception(Exception::StorePageFault) => {
            // 判断是否是lazy
            let va = VirtualAddr(stval::read());
            let pcb = current_pcb().unwrap();
            let mut pcblock = pcb.lock();
            if let Ok(_) = pcblock
                .memory_space
                .mmap_areas
                .check_lazy(va, MapProt::WRITE)
            {
                // 已经分配物理页，由current_hart_run映射。
                log!("mmap":"store">"Found mapped page va(0x{:x})", va.0);
                drop(pcblock);
                current_hart_leak();
                scheduler_insert_front(pcb);
                schedule();
            } else if let Ok(_) = pcblock.memory_space.prog_check_lazy(va) {
                drop(pcblock);
                current_hart_leak();
                scheduler_insert_front(pcb);
                schedule();
            } else {
                log!("mmap":"store">"Not Found mapped page va(0x{:x})", va.0);
                pcblock.exit(-1);
                current_hart_leak();
                schedule();
            }
        }
        Trap::Exception(Exception::LoadFault) | Trap::Exception(Exception::LoadPageFault) => {
            // 判断是否是lazy
            let va = VirtualAddr(stval::read());
            let pcb = current_pcb().unwrap();
            let mut pcblock = pcb.lock();
            if let Ok(_) = pcblock
                .memory_space
                .mmap_areas
                .check_lazy(va, MapProt::READ)
            {
                // 已经分配物理页，由current_hart_run映射。
                log!("mmap":"load">"Found mapped page va(0x{:x})", va.0);
                drop(pcblock);
                current_hart_leak();
                scheduler_insert_front(pcb);
                schedule();
            } else if let Ok(_) = pcblock.memory_space.prog_check_lazy(va) {
                drop(pcblock);
                current_hart_leak();
                scheduler_insert_front(pcb);
                schedule();
            } else {
                log!("mmap":"load">"Not Found mapped page va(0x{:x})", va.0);
                pcblock.exit(-1);
                current_hart_leak();
                schedule();
            }
        }
        Trap::Exception(Exception::IllegalInstruction) => {
            // schedule();
            let sepc = current_pcb().unwrap().lock().trapframe()["sepc"];
            let instructions = VirtualAddr(sepc);
            let pte = current_hart_pgtbl().walk(instructions, false);
            log!("test_inst_p_f":"page fault">"sepc {:#x}, ppn {:#x}, flags {:?}", sepc, pte.ppn().page(), pte.flags());
            let instructions: PhysAddr = instructions.into();
            let instructions: &[u8; 6] = instructions.as_ref();
            println!(
                "IllegalInstructoin, core dumped, sepc: 0x{:x}, scause:{:?}, instructions is [{:#x},{:#x},{:#x},{:#x},{:#x},{:#x}] ",
                sepc,
                scause.cause(),
                instructions[0], instructions[1],instructions[2],instructions[3],instructions[4],instructions[5],
            );
            current_pcb().unwrap().lock().exit(-1);
            current_hart_leak();
            schedule();
        }
        Trap::Exception(Exception::InstructionPageFault) => {
            // schedule();
            let sepc = current_pcb().unwrap().lock().trapframe()["sepc"];
            let instructions = VirtualAddr(sepc);
            let pte = current_hart_pgtbl().walk(instructions, false);
            log!("test_inst_p_f":"page fault">"sepc {:#x}, ppn {:#x}, flags {:?}", sepc, pte.ppn().page(), pte.flags());
            let instructions: PhysAddr = instructions.into();
            let instructions: &[u8; 6] = instructions.as_ref();
            println!(
                "InstructionPageFault, core dumped, sepc: 0x{:x}, scause:{:?}, instructions is [{:#x},{:#x},{:#x},{:#x},{:#x},{:#x}] ",
                sepc,
                scause.cause(),
                instructions[0], instructions[1],instructions[2],instructions[3],instructions[4],instructions[5],
            );
            current_pcb().unwrap().lock().exit(-1);
            current_hart_leak();
            schedule();
        }
        Trap::Interrupt(Interrupt::SupervisorTimer) => {
            log!("trap":"time_interrupt">"");
            hart_set_next_trigger();
            let pcb = current_pcb().unwrap();
            current_hart_leak();
            scheduler_insert_front(pcb);
            schedule();
        }
        Trap::Exception(Exception::Breakpoint) => {
            let pcb = current_pcb().unwrap();
            let mut pcblock = pcb.lock();
            pcblock.trapframe()["sepc"] += 4;
            drop(pcblock);
            current_hart_leak();
            scheduler_insert_front(pcb);
            schedule();
        }
        _ => {
            // schedule();
            panic!(
                "Unsupported trap {:?}:0x{:x}, sepc = {:#x} stval = {:#x}!",
                scause.cause(),
                scause.bits(),
                current_pcb().unwrap().lock().trapframe()["sepc"],
                stval
            );
        }
    };
}
