use super::*;
use alloc::vec::Vec;
use alloc::collections::BTreeMap;
use spin::Mutex;
use spin::Lazy;
use super::Ino;

static MOUNT_POINTS: Lazy<Mutex<BTreeMap<Ino, Inode>>> = Lazy::new(|| { Mutex::new(BTreeMap::new()) });

pub fn split_path<'c>(path: &'c str) -> (&'c str, Option<&'c str>) {
    let mut path_split = path.trim_matches('/').splitn(2, '/');
    let comp = path_split.next().unwrap(); // SAFE: splitn always returns at least one element
    let rest_opt = path_split.next();
    (comp, rest_opt)
}

pub fn rsplit_path<'c>(path: &'c str) -> (Option<&'c str>, &'c str) {
    let mut path_split = path.trim_end_matches('/').rsplitn(2, '/');
    let comp = path_split.next().unwrap(); // SAFE: splitn always returns at least one element
    let rest_opt = path_split.next();
    (rest_opt, comp)
}

pub fn parse_path(from: &Inode, path: &str) -> Result<Inode, FileErr> {
    log!("path_resolve":"{}">"", path);
    let mut nodes = Vec::new();
    nodes.push(from.clone());
    let mut rest = Some(path);
    while let Some(rest_path) = rest {
        if let Some(mut inode) = nodes.last().cloned() {
            // resolve mount point
            if let Ok(mut ino) = inode.ino() {
                log!("path_resolve":"mountpoint">"test ino {}", ino);
                if let Some(source) = MOUNT_POINTS.lock().get(&ino) {
                    log!("path_resolve":"mountpoint">"ino {} is a mount point", ino);
                    inode = source.clone();
                }
            }
            let (name, mut _rest) = split_path(rest_path);
            rest = _rest;

            log!("path_resolve":>"item name {}", name);
            if name.len() == 0 {
                // Continue
            } else if name == "." {
                continue;
            } else if name == ".." {
                nodes.pop();
                continue;
            } else {
                let child = inode.get_child(name)?;
                nodes.push(child);
            }
        } else {
            // ".." 超过根目录，比如"/dir/../.."
            return Err(FileErr::NotDefine);
        }
    }
    match nodes.last() {
        Some(inode) => {
            log!("path_resolve":"success">"{}", path);
            Ok(inode.clone())
        }
        None => Err(FileErr::NotDefine),
    }
}

pub fn is_absolute_path(path: &str) -> bool {
    match path.chars().next() {
        Some('/') => true,
        _ => false,
    }
}

pub fn mount_at(source: Inode, target: Ino) -> Result<(), FileErr> {
    if MOUNT_POINTS.lock().contains_key(&target) {
        log!("sys_mount":"mount_at">"already mount {}", target);
        return Err(FileErr::NotDefine)
    }
    MOUNT_POINTS.lock().insert(target, source);
    Ok(())
}

pub fn umount_at(target: Ino) -> Result<(), FileErr> {
    match MOUNT_POINTS.lock().remove(&target) {
        Some(_) => {
            Ok(())
        }
        None => {
            Err(FileErr::NotDefine)
        }
    }
}
