use super::*;
use alloc::sync::Arc;
use spin::Mutex;

#[derive(Default)]
struct PipeInode {
    inner: Mutex<PipeInner>,
}

const PIPE_INODE_SIZE: usize = 512;
struct PipeInner {
    read_ready: bool,
    write_ready: bool,
    // 记录有几个文件读该Inode
    reader: usize,
    // 记录有几个文件写该Inode
    writer: usize,
    // 数据开始
    begin: usize,
    // 数据结束
    end: usize,
    // 数据结束
    data: [u8; PIPE_INODE_SIZE],
}

impl Default for PipeInner {
    fn default() -> Self {
        Self {
            read_ready: true,
            write_ready: true,
            reader: 0,
            writer: 0,
            begin: 0,
            end: 0,
            data: [0; PIPE_INODE_SIZE],
        }
    }
}

impl PipeInner {
    pub fn len(&self) -> usize {
        usize::MAX
    }

    pub fn empty(&self) -> bool {
        self.begin == self.end
    }

    pub fn full(&self) -> bool {
        (self.end + 1) % PIPE_INODE_SIZE == self.begin
    }

    pub fn contains_data(&self) -> usize {
        (self.end + PIPE_INODE_SIZE - self.begin) % PIPE_INODE_SIZE
    }

    pub fn remaining_space(&self) -> usize {
        PIPE_INODE_SIZE - self.contains_data() - 1
    }
}

impl _Inode for PipeInode {
    fn len(&self) -> usize {
        usize::MAX
    }

    fn read_offset(&self, _: usize, buf: &mut [u8]) -> Result<usize, FileErr> {
        // 管道读时忽略offset参数
        log!("pipe":"read_offset">"len ({})", buf.len());
        let mut inner = self.inner.lock();

        if inner.empty() {
            // 如果管道为空
            if inner.writer == 0 {
                // 没有写者了
                return Err(FileErr::FileEOF);
            } else {
                // 还有写者，还有写入的机会
                return Err(FileErr::PipeReadWait);
            }
        }

        // 本次读取的长度由pipe内缓存的数据与缓冲区大小决定
        let read_len = core::cmp::min(inner.contains_data(), buf.len());
        if inner.end > inner.begin {
            // 缓冲区位置连续
            // [......*******.....]
            //  begin ^      ^ end
            
            // 直接拷贝
            buf[0..read_len].copy_from_slice(&inner.data[inner.begin..inner.begin+read_len]);
        } else {
            // 缓冲区位置不连续，可能要分两次拷贝
            // [****.......*******]
            //  end ^      ^ begin
            
            // 第一次最多从begin拷贝到pipe缓冲区末尾
            let first_step = core::cmp::min(read_len, PIPE_INODE_SIZE-inner.begin);
            // 第一次最多从0拷贝到end
            let second_step = if read_len <= first_step { first_step } else { read_len - first_step };

            buf[0..first_step].copy_from_slice(&inner.data[inner.begin..inner.begin+first_step]);
            buf[first_step..second_step].copy_from_slice(&inner.data[0..second_step-first_step]);
        }

        // 更新begin指针
        inner.begin = (inner.begin + read_len) % PIPE_INODE_SIZE;
        return Ok(read_len);
    }

    fn write_offset(&self, _: usize, buf: &[u8]) -> Result<usize, FileErr> {
        // 管道写时忽略offset参数
        log!("pipe":"write_offset">"len ({})", buf.len());
        let mut inner = self.inner.lock();

        if inner.full() {
            // 如果管道已满
            if inner.reader == 0 {
                // 没有读者了
                return Err(FileErr::FileEOF);
            } else {
                // 还有读者，还有读取的机会
                return Err(FileErr::PipeWriteWait);
            }
        }

        // 本次写入的长度由pipe内剩余空间与缓冲区大小决定
        let write_len = core::cmp::min(inner.remaining_space(), buf.len());
        if inner.begin > inner.end {
            // 剩余空间位置连续
            // [****.......*******]
            //  end ^      ^ begin
            
            // 直接拷贝
            let end = inner.end;
            inner.data[end..end+write_len].copy_from_slice(&buf[0..write_len]);
        } else {
            // 剩余空间位置连续，可能要分两次拷贝
            // [......*******.....]
            //  begin ^      ^ end
            
            // 第一次最多从end拷贝到pipe缓冲区末尾
            let first_step = core::cmp::min(write_len, PIPE_INODE_SIZE-inner.end);
            // 第一次最多从0拷贝到begin
            let second_step = if write_len <= first_step { first_step } else { write_len - first_step };

            let end = inner.end;
            inner.data[end..end+first_step].copy_from_slice(&buf[0..first_step]);
            inner.data[0..second_step-first_step].copy_from_slice(&buf[first_step..second_step]);
        }

        // 更新end指针
        inner.end = (inner.end + write_len) % PIPE_INODE_SIZE;
        return Ok(write_len);
    }

    fn file_open(&self, flags: OpenFlags) {
        let mut inner = self.inner.lock();
        if flags.readable() {
            inner.reader += 1;
        }
        if flags.writable() {
            inner.writer += 1;
        }
        log!("pipe":"file_open">"remain reader({}), remain writer({})", inner.reader, inner.writer);
    }

    fn file_close(&self, file: &File) {
        let mut inner = self.inner.lock();
        if file.flags().readable() {
            inner.reader -= 1;
        }
        if file.flags().writable() {
            inner.writer -= 1;
        }
        log!("pipe":"file_close">"remain reader({}), remain writer({})", inner.reader, inner.writer);
    }

    fn read_ready(&self) -> bool {
        let inner = self.inner.lock();
        // 如果没有写入者时可以read
        if inner.writer == 0 {
            true
        } else {
            inner.read_ready
        }
    }

    fn write_ready(&self) -> bool {
        let inner = self.inner.lock();
        // 如果没有读出者时可以write
        if inner.reader == 0 {
            true
        } else {
            inner.write_ready
        }
    }
}

pub fn make_pipe() -> Result<(Fd, Fd), FileErr> {
    let pipe = Arc::new(PipeInode::default());
    let reader = File::open(pipe.clone(), OpenFlags::RDONLY)?;
    let writer = File::open(pipe, OpenFlags::WRONLY)?;
    Ok((reader, writer))
}
